import psutil
import dask
import numpy as np
import argparse
import traceback
import time
import sys
import os
import logging
from dask import delayed
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.ERROR)
logging.getLogger("tornado.application").setLevel(logging.CRITICAL)
datadir = get_datadir()


def get_polmodel_coeff(model_file):
    """
    Get polarization model coefficients

    Parameters
    ----------
    model_file : str
        Model ascii file name

    Returns
    -------
    float
        Reference frequency in GHz
    float
        Stokes I flux density in Jy at reference frequency
    list
        Stokes I spectrum log-polynomial coefficients
    list
        List of linear polarization fraction polynomial coefficients
    list
        List of linear polarization angle (in radians) coefficients
    """
    freq, I, pfrac, pangle = np.loadtxt(model_file, unpack=True, skiprows=5)
    ref_freq = freq[0]
    x = (freq - ref_freq) / ref_freq
    polyI = np.polyfit(np.log10(x[1:]), np.log10(I[1:]), deg=5)[::-1]
    poly_pfrac = np.polyfit(x, pfrac, deg=5)[::-1]
    poly_pangle = np.polyfit(x, pangle, deg=5)[::-1]
    return ref_freq, I[0], polyI[1:], poly_pfrac, poly_pangle


def polcal_setjy(msname="", field_name="", n_threads=-1, ismms=True):
    """
    Setjy polcal fields (3C286 or 3C138)

    Parameters
    ----------
    msname : str
        Name of measurement set
    field : str, optional
        Field name
    n_threads : int, optional
        Number of OpenMP threads
    """
    limit_threads(n_threads=n_threads)
    from casatasks import setjy

    try:
        print(f"Polarization calibrator field name : {field_name}")
        if field_name in ["3C286", "1328+307", "1331+305", "J1331+3030"]:
            ref_freq, I, polyI, poly_pfrac, poly_pangle = get_polmodel_coeff(
                datadir + "/3C286_pol_model.txt"
            )
        elif field_name in ["3C138", "0518+165", "0521+166", "J0521+1638"]:
            ref_freq, I, polyI, poly_pfrac, poly_pangle = get_polmodel_coeff(
                datadir + "/3C138_pol_model.txt"
            )
        else:
            print("Field name is not either of the polcal, 3C286 or 3C138.")
            return 1
        print(f"Reference frequency: {ref_freq} GHz.")
        with suppress_output():
            setjy(
                vis=msname,
                field=field_name,
                scalebychan=True,
                standard="manual",
                fluxdensity=[I, 0.0, 0.0, 0.0],
                spix=polyI,
                reffreq=f"{ref_freq}GHz",
                polindex=poly_pfrac,
                polangle=poly_pangle,
                rotmeas=0,
                usescratch=True,
            )
    except Exception as e:
        traceback.print_exc()
    return


def phasecal_setjy(msname="", field="", ismms=False, n_threads=-1):
    """
    Setjy phasecal fields

    Parameters
    ----------
    msname : str
        Measurement set
    field : str, optional
        Phasecal fields
    ismms : bool, optional
        Is a multi-ms ot not
    n_threads : int, optional
        Number of OpenMP threads
    """
    limit_threads(n_threads=n_threads)
    from casatasks import setjy

    try:
        with suppress_output():
            setjy(
                vis=msname,
                field=field,
                standard="manual",
                fluxdensity=[1.0, 0, 0, 0],
                usescratch=True,
                ismms=ismms,
            )
    except Exception as e:
        traceback.print_exc()
    return


def import_fluxcal_models(mslist, fluxcal_fields, fluxcal_scans, ncpus=1, mem_frac=0.8):
    """
    Import model visibilities using crystalball

    Parameters
    ----------
    mslist : list
        List of sub-MS
    fluxcal_fields : list
        Fluxcal fields
    fluxcal_scans : dict
        Fluxcal scans dict
    ncpus : int
        Number of CPU threads to use
    mem_frac : float
        Memory fraction to use

    Returns
    -------
    int
        Success message
    """
    try:
        if len(fluxcal_fields) == 0:
            print("No flux calibrator scan is present.")
            return 1
        print("##############################")
        print("Import fluxcal models")
        print("##############################")
        bandname = get_band_name(mslist[0])
        cpu_count = psutil.cpu_count()
        ncpus = max(1, ncpus)
        ncpus = min(ncpus, cpu_count - 1)
        mem_frac = max(mem_frac, 0.8)
        scans = []
        for ms in mslist:
            ms_scans = get_ms_scans(ms)
            for s in ms_scans:
                scans.append(s)
        for fluxcal in fluxcal_fields:
            f_scan = fluxcal_scans[fluxcal]
            for s in f_scan:
                if s in scans:
                    pos = scans.index(s)
                    sub_msname = mslist[pos]
                    modelname = datadir + "/" + fluxcal + "_" + bandname + "_model.txt"
                    crys_cmd_args = [
                        "-sm " + modelname,
                        "-f " + fluxcal,
                        "-ns 1000",
                        "-j " + str(ncpus),
                        "-mf " + str(mem_frac),
                        sub_msname,
                    ]
                    cmd = ["crystalball"] + crys_cmd_args
                    cmd = " ".join(cmd)
                    print(f"Running: {cmd}")
                    tmpfile = f"tmp_{os.path.basename(sub_msname)}"
                    with suppress_output():
                        msg = os.system(f"{cmd} > {tmpfile}")
                    if os.path.exists(tmpfile):
                        os.remove(tmpfile)
                    if msg == 0:
                        print(f"Fluxcal model is imported successfully for scan: {s}.")
                    else:
                        print(f"Error in importing fluxcal model for scan: {s}.")
        return 0
    except Exception as e:
        print("Error in importing model.")
        traceback.print_exc()
        return 1


def import_phasecal_models(
    mslist,
    dask_client,
    phasecal_fields,
    phasecal_scans,
    workdir,
    cpu_frac=0.8,
):
    """
    Import model visibilities for phasecal

    Parameters
    ----------
    mslist : list
        List of sub-MS
    dask_client : dask.client
        Dask client
    phasecal_fields : list
        Phasecal fields
    phasecal_scans : list
        Phasecal scans
    workdir : str
        Work directory
    cpu_frac : float, optional
        CPU fraction to use

    Returns
    -------
    int
        Success message
    """
    try:
        if cpu_frac > 0.8:
            cpu_frac = 0.8
        total_cpu = max(1, int(psutil.cpu_count() * cpu_frac))
        print("##########################")
        print("Import phasecal models")
        print("##########################")
        scans = []
        for ms in mslist:
            ms_scans = get_ms_scans(ms)
            for s in ms_scans:
                scans.append(s)
        tasks = []
        total_scans = 0
        for phasecal in phasecal_fields:
            ph_scan = phasecal_scans[phasecal]
            for s in ph_scan:
                if s in scans:
                    total_scans += 1
        n_threads = max(1, int(total_cpu / total_scans))
        for phasecal in phasecal_fields:
            ph_scan = phasecal_scans[phasecal]
            for s in ph_scan:
                if s in scans:
                    pos = scans.index(s)
                    sub_msname = mslist[pos]
                    tasks.append(
                        delayed(phasecal_setjy)(
                            sub_msname,
                            field=phasecal,
                            ismms=True,
                            n_threads=n_threads,
                        )
                    )
        print("Start importing phasecal models...")
        results = list(dask_client.gather(dask_client.compute(tasks)))
        print("Phasecal models are initiated.")
        return 0
    except Exception as e:
        print("Error in initiaing phasecal models.")
        traceback.print_exc()
        return 1


def import_polcal_models(
    mslist,
    dask_client,
    polcal_fields,
    polcal_scans,
    workdir,
    cpu_frac=0.8,
):
    """
    Import model for polarization calibrators (3C286 or 3C138)

    Parameters
    ----------
    mslist : list
        List of sub-MS
    dask_client : dask.client
        Dask client
    polcal_fields : list
        Polcal fields
    polcal_scans : list
        Polcal scans
    workdir : str
        Work directory
    cpu_frac : float, optional
        CPU fraction to use

    Returns
    -------
    int
        Success message
    """
    try:
        if cpu_frac > 0.8:
            cpu_frac = 0.8
        total_cpu = max(1, int(psutil.cpu_count() * cpu_frac))
        if len(polcal_scans) == 0:
            print("No polarization calibrator scans is present.")
            return 1
        scans = []
        for ms in mslist:
            ms_scans = get_ms_scans(ms)
            for s in ms_scans:
                scans.append(s)
        tasks = []
        total_scans = 0
        for polcal_field in polcal_fields:
            p_scan = polcal_scans[polcal_field]
            for s in p_scan:
                if s in scans:
                    total_scans += 1
        n_threads = max(1, int(total_cpu / total_scans))
        for polcal_field in polcal_fields:
            p_scan = polcal_scans[polcal_field]
            for s in p_scan:
                if s in scans:
                    pos = scans.index(s)
                    sub_msname = mslist[pos]
                    tasks.append(
                        delayed(polcal_setjy)(
                            sub_msname, polcal_field, ismms=True, n_threads=n_threads
                        )
                    )
        print("Start importing polcal models...")
        results = list(dask_client.gather(dask_client.compute(tasks)))
        print("Polcal models imports are done.")
        return 0
    except Exception as e:
        traceback.print_exc()
        return 1


def import_all_models(msname, dask_client, workdir, cpu_frac=0.8, mem_frac=0.8):
    """
    Import all calibrator models

    Parameters
    ----------
    msname : str
        Measurement set name
    dask_client : dask.client
        Dask client
    workdir : str
        Work directory
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use

    Returns
    -------
    int
        Success message
    """
    cpu_threads = psutil.cpu_count()
    ncpu = int(cpu_threads * cpu_frac)
    mem_frac = 1 - mem_frac
    try:
        msname = msname.rstrip("/")
        correct_missing_col_subms(msname)
        mspath = os.path.dirname(os.path.abspath(msname))
        os.chdir(mspath)
        result = get_submsname_scans(msname)
        if result is not None:  # If multi-ms
            mslist, scans = result
        else:
            print("Please provide a multi-MS with scans partitioned.")
            return 1, 1, 1
        fluxcal_fields, fluxcal_scans = get_fluxcals(msname)
        phasecal_fields, phasecal_scans, phasecal_flux_list = get_phasecals(msname)
        polcal_fields, polcal_scans = get_polcals(msname)
        fluxcal_result = import_fluxcal_models(
            mslist, fluxcal_fields, fluxcal_scans, ncpus=ncpu, mem_frac=mem_frac
        )
        if fluxcal_result != 0:
            return fluxcal_result, 1, 1

        phasecal_result = import_phasecal_models(
            mslist,
            dask_client,
            phasecal_fields,
            phasecal_scans,
            workdir,
            cpu_frac=cpu_frac,
        )

        polcal_result = import_polcal_models(
            mslist,
            dask_client,
            polcal_fields,
            polcal_scans,
            workdir,
            cpu_frac=cpu_frac,
        )

        if phasecal_result != 0:
            print(
                "Phasecal model was not import, but fluxcal model import is successful."
            )

        if polcal_result != 0:
            print(
                "Polcal model was not imported, but fluxcal model import is successful."
            )
        return fluxcal_result, phasecal_result, polcal_result
    except Exception as e:
        traceback.print_exc()
        return 1, 1, 1


def main(
    msname,
    workdir="",
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid=0,
    start_remote_log=False,
    dask_client=None,
):
    """
    Run the calibrator model import pipeline.

    Parameters
    ----------
    msname : str
        Path to the measurement set (MS) to process.
    workdir : str, optional
        Working directory for intermediate and output files. If empty, defaults
        to `<msname>/workdir`. Default is "".
    cpu_frac : float, optional
        Fraction of total CPU cores to use. Default is 0.8.
    mem_frac : float, optional
        Fraction of total system memory to use. Default is 0.8.
    logfile : str or None, optional
        Path to the logfile for logging messages. If None, logging to file is disabled.
    jobid : int, optional
        Job identifier for tracking and PID management. Default is 0.
    start_remote_log : bool, optional
        If True, enables remote logging based on credentials stored in the
        working directory. Default is False.
    dask_client : dask.client, optional
        Dask client

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "import_model", logfile, jobname=jobname, password=password
            )

    dask_cluster = None
    if dask_client is None:
        dask_client, dask_cluster, dask_dir = get_local_dask_cluster(
            2,
            dask_dir=workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
        )
        nworker = max(2, int(psutil.cpu_count() * cpu_frac))
        scale_worker_and_wait(dask_cluster, nworker)

    try:
        if msname and os.path.exists(msname):
            fluxcal_result, phasecal_result, polcal_result = import_all_models(
                os.path.abspath(msname),
                dask_client,
                workdir,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
            )
            os.system("touch " + workdir + "/.fluxcal_" + str(fluxcal_result))
            os.system("touch " + workdir + "/.phasecal_" + str(phasecal_result))
            os.system("touch " + workdir + "/.polcal_" + str(polcal_result))
            if fluxcal_result == 1 or (
                fluxcal_result == 1 and phasecal_result == 1 and polcal_result == 1
            ):
                msg = 1
            else:
                msg = 0
        else:
            print("Please provide correct measurement set.")
            msg = 1
    except Exception as e:
        traceback.print_exc()
        os.system("touch " + workdir + "/.fluxcal_1")
        os.system("touch " + workdir + "/.phasecal_1")
        os.system("touch " + workdir + "/.polcal_1")
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)
        clean_shutdown(observer)
        if dask_cluster is not None:
            dask_client.close()
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
    return msg


def cli():
    usage = "Import calibrator models"
    parser = argparse.ArgumentParser(
        description=usage, formatter_class=SmartDefaultsHelpFormatter
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument("msname", type=str, help="Name of measurement set")
    basic_args.add_argument(
        "--workdir", type=str, default="", help="Name of work directory"
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac", type=float, default=0.8, help="CPU fraction to use"
    )
    hard_args.add_argument(
        "--mem_frac", type=float, default=0.8, help="Memory fraction to use"
    )
    hard_args.add_argument("--logfile", type=str, default=None, help="Log file")
    hard_args.add_argument("--jobid", type=int, default=0, help="Job ID")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        return 1

    args = parser.parse_args()

    msg = main(
        msname=args.msname,
        workdir=args.workdir,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nVisibility simulation is finished.\n###################\n"
    )
    os._exit(result)



import logging
import dask
import numpy as np
import argparse
import traceback
import warnings
import time
import glob
import sys
import os
from casatools import msmetadata, ms as casamstool
from dask import delayed
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.ERROR)
logging.getLogger("tornado.application").setLevel(logging.CRITICAL)
datadir = get_datadir()


def make_solar_DS(
    msname,
    dask_client,
    workdir,
    ds_file_name="",
    extension="png",
    target_scans=[],
    scans=[],
    merge_scan=False,
    showgui=False,
    cpu_frac=0.8,
    mem_frac=0.8,
):
    """
    Make solar dynamic spectrum and plots

    Parameters
    ----------
    msname : str
        Measurement set name
    dask_client : dask.client
        Dask client
    workdir : str
        Work directory
    ds_file_name : str, optional
        DS file name prefix
    extension : str, optional
        Image file extension
    target_scans : list, optional
        Target scans
    scans : list, optional
        Scan list
    merge_scan : bool, optional
        Merge scans in one plot or not
    showgui : bool, optional
        Show GUI
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    """
    if cpu_frac > 0.8:
        cpu_frac = 0.8
    total_cpu = max(1, int(psutil.cpu_count() * cpu_frac))
    if mem_frac > 0.8:
        mem_frac = 0.8
    total_mem = (psutil.virtual_memory().available * mem_frac) / (1024**3)  # In GB

    warnings.filterwarnings("ignore", category=RuntimeWarning)
    os.makedirs(f"{workdir}/dynamic_spectra", exist_ok=True)
    print("##############################################")
    print(f"Start making dynamic spectra for ms: {msname}")
    print("##############################################")
    if len(target_scans) > 0:
        temp_target_scans = []
        for s in target_scans:
            temp_target_scans.append(int(s))
        target_scans = temp_target_scans

    ##################################
    # Making and ploting
    ##################################
    if len(scans) == 0:
        scans, cal_scans, f_scans, g_scans, p_scans = get_cal_target_scans(msname)
    valid_scans = get_valid_scans(msname)
    final_scans = []
    scan_size_list = []
    msmd = msmetadata()
    mstool = casamstool()
    for scan in scans:
        if scan in valid_scans:
            if len(target_scans) == 0 or (
                len(target_scans) > 0 and int(scan) in target_scans
            ):
                final_scans.append(int(scan))
                scan_size = get_ms_scan_size(msname, int(scan), only_autocorr=True)
                scan_size_list.append(scan_size)
    if len(final_scans) == 0:
        print("No scans to make dynamic spectra.")
        return
    del scans
    scans = sorted(final_scans)
    print(f"Scans: {scans}")
    msname = msname.rstrip("/")
    if ds_file_name == "":
        ds_file_name = os.path.basename(msname).split(".ms")[0] + "_DS"
    hascor = check_datacolumn_valid(msname, datacolumn="CORRECTED_DATA")
    if hascor:
        datacolumn = "CORRECTED_DATA"
    else:
        datacolumn = "DATA"
    mspath = os.path.dirname(msname)

    ########################################
    # Number of worker limit based on memory
    ########################################
    mem_limit = min(total_mem, max(scan_size_list))
    njobs = min(len(scans), int(total_mem / mem_limit))
    njobs = max(1, min(total_cpu, njobs))
    n_threads = max(1, int(total_cpu / njobs))

    print("#################################")
    print(f"Total dask worker: {njobs}")
    print(f"CPU per worker: {n_threads}")
    print(f"Memory per worker: {round(mem_limit,2)} GB")
    print("#################################")
    ###########################################

    tasks = []
    for scan in scans:
        tasks.append(
            delayed(make_ds_file_per_scan)(
                msname,
                f"{workdir}/dynamic_spectra/{ds_file_name}_scan_{scan}",
                scan,
                datacolumn,
            )
        )
    results = []
    print("Start making dynamic spectra...")
    for i in range(0, len(tasks), njobs):
        batch = tasks[i : i + njobs]
        futures = dask_client.compute(batch)
        results.extend(dask_client.gather(futures))
    results = list(results)

    ds_files = [
        f"{workdir}/dynamic_spectra/{ds_file_name}_scan_{scan}.npy" for scan in scans
    ]
    print(f"DS files: {ds_files}")
    plots = []
    for dsfile in ds_files:
        plot_file = make_ds_plot(
            [dsfile],
            plot_file=dsfile.replace(".npy", f".{extension}"),
            showgui=showgui,
        )
        plots.append(plot_file)
    if merge_scan:
        plot_file = make_ds_plot(
            ds_files,
            plot_file=f"{workdir}/dynamic_spectra/{ds_file_name}.{extension}",
            showgui=showgui,
        )
    goes_files = glob.glob(f"{workdir}/dynamic_spectra/sci*.nc")
    for f in goes_files:
        os.system(f"rm -rf {f}")
    return


def make_dsfiles(
    msname,
    dask_client,
    workdir,
    outdir,
    extension="png",
    target_scans=[],
    seperate_scans=True,
    cpu_frac=0.8,
    mem_frac=0.8,
):
    """
    Make all dynamic spectra of the solar scans

    Parameters
    ----------
    msname : str
        Measurement set name
    dask_client : dask.client
        Dask client
    workdir : str
        Work directory
    outdir : str
        Output directory
    extension : str, optional
        Plot file extension
    target_scans : list, optional
        Target scans
    seperate_scans : bool, optional
        Only seperate scan plots
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use

    Returns
    -------
    list
        Dynamic spectra files
    """
    msname = msname.rstrip("/")
    workdir = workdir.rstrip("/")
    if seperate_scans == False:
        return
    try:
        if seperate_scans:
            make_solar_DS(
                msname,
                dask_client,
                workdir,
                extension=extension,
                target_scans=target_scans,
                merge_scan=False,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
            )
        else:
            make_solar_DS(
                msname,
                dask_client,
                workdir,
                extension=extension,
                target_scans=target_scans,
                merge_scan=True,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
            )
        if os.path.samefile(outdir, workdir) == False:
            os.makedirs(f"{outdir}/dynamic_spectra", exist_ok=True)
            os.system(f"mv {workdir}/dynamic_spectra/* {outdir}/dynamic_spectra/")
            os.system(f"rm -rf {workdir}/dynamic_spectra")
        ds_file_name = os.path.basename(msname).split(".ms")[0] + "_DS"
        ds_files = glob.glob(f"{outdir}/dynamic_spectra/{ds_file_name}*.{extension}")
        return ds_files
    except Exception as e:
        traceback.print_exc()
        return []
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)


def main(
    msname,
    workdir,
    outdir,
    extension="png",
    target_scans=[],
    seperate=True,
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid="0",
    start_remote_log=False,
    dask_client=None,
):
    """
    Make dynamic spectra

    Parameters
    ----------
    msname : str
        Measurement set
    workdir : str
        Work directory
    outdir : str
        Output directory
    extension : str, optional
        Plot extension
    target_scans : list, optional
        Target scans
    seperate : bool, optional
        Only seperate scan plots
    cpu_frac : float, optional
        CPU fraction
    mem_frac : float, optional
        Memory fraction
    logfile : str, optional
        Log file
    jobid : str, optional
        Job ID
    start_remote_log : bool, optional
        Start remote log
    dask_client: dask.client, optional
        Dask client

    Returns
    -------
    int
        Success messsage
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "ds_plot", logfile, jobname=jobname, password=password
            )

    dask_cluster = None
    if dask_client is None:
        dask_client, dask_cluster, dask_dir = get_local_dask_cluster(
            2,
            dask_dir=workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
        )
        nworker = max(2, int(psutil.cpu_count() * cpu_frac))
        scale_worker_and_wait(dask_cluster, nworker)

    try:
        if msname != "" and os.path.exists(msname):
            ds_files = make_dsfiles(
                msname,
                dask_client,
                workdir,
                outdir,
                extension=extension,
                target_scans=target_scans,
                seperate_scans=seperate,
                cpu_frac=float(cpu_frac),
                mem_frac=float(mem_frac),
            )
            msg = 0
        else:
            print("Please provide a valid measurement set.")
            msg = 1
    except Exception as e:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)
        clean_shutdown(observer)
        if dask_cluster is not None:
            dask_client.close()
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Make dynamic spectra of solar scans",
        formatter_class=SmartDefaultsHelpFormatter,
    )
    # === Essential parameters ===
    essential = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    essential.add_argument("msname", type=str, help="Measurement set name")
    essential.add_argument(
        "--workdir",
        type=str,
        dest="workdir",
        required=True,
        help="Working directory",
    )
    essential.add_argument(
        "--outdir",
        type=str,
        dest="outdir",
        required=True,
        help="Output directory",
    )

    # === Advanced parameters ===
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--extension",
        type=str,
        default="png",
        help="Save file extension",
    )
    adv_args.add_argument(
        "--target_scans",
        nargs="*",
        type=str,
        default=[],
        help="List of target scans to process (space-separated, e.g. 3 5 7)",
    )
    adv_args.add_argument(
        "--no_seperate",
        action="store_false",
        dest="seperate",
        help="Do not seperate scans",
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # === Advanced local system/ per node hardware resource parameters ===
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="Fraction of CPU usuage per node",
    )
    hard_args.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Fraction of memory usuage per node",
    )
    hard_args.add_argument(
        "--logfile", type=str, default=None, help="Optional path to log file"
    )
    hard_args.add_argument(
        "--jobid", type=str, default="0", help="Job ID for logging and PID tracking"
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        return 1

    args = parser.parse_args()

    msg = main(
        msname=args.msname,
        workdir=args.workdir,
        outdir=args.outdir,
        extension=args.extension,
        target_scans=args.target_scans,
        seperate=args.seperate,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nDynamic spectra are produced successfully.\n###################\n"
    )
    os._exit(result)



import os
import types
import subprocess
import time
import socket
import signal
import argparse
import toml
from pathlib import Path
from dotenv import load_dotenv
from meersolar.utils.basic_utils import *


# === CONFIG ===
def prefect_config():
    cachedir = get_cachedir()
    PREFECT_HOME = f"{cachedir}/prefect_home"
    os.makedirs(PREFECT_HOME, exist_ok=True)

    DB_URL = f"sqlite+aiosqlite:///{PREFECT_HOME}/prefect.db"
    LOG_FILE = os.path.join(PREFECT_HOME, "server.log")
    profile_path = os.path.join(PREFECT_HOME, "profiles.toml")
    memo_path = os.path.join(PREFECT_HOME, "memo_store.toml")
    storage = os.path.join(PREFECT_HOME, "storage")
    os.makedirs(storage, exist_ok=True)
    ENV_FILE = os.path.join(cachedir, "meersolar_prefect.env")
    SERVER_HOST = "0.0.0.0"
    SERVER_PORT = "4250"
    SERVER_URL = f"http://{SERVER_HOST}:{SERVER_PORT}/api"
    SERVER_DASHBOARD = f"http://{SERVER_HOST}:{SERVER_PORT}/dashboard"
    profile_name = "solarpipe"
    pid_file = os.path.join(PREFECT_HOME, "server.pid")
    logging_path = os.path.join(PREFECT_HOME, "logging.yml")

    return {
        "CACHEDIR": cachedir,
        "PREFECT_HOME": PREFECT_HOME,
        "DB_URL": DB_URL,
        "LOG_FILE": LOG_FILE,
        "PROFILE_PATH": profile_path,
        "MEMO_PATH": memo_path,
        "STORAGE": storage,
        "ENV_FILE": ENV_FILE,
        "SERVER_HOST": SERVER_HOST,
        "SERVER_PORT": SERVER_PORT,
        "SERVER_URL": SERVER_URL,
        "SERVER_DASHBOARD": SERVER_DASHBOARD,
        "PROFILE_NAME": profile_name,
        "PID_FILE": pid_file,
        "LOGGING_PATH": logging_path,
    }


def write_prefect_profile():
    """
    Save prefect profile
    """
    config = prefect_config()
    # Load existing TOML config or start new
    profile_path = config["PROFILE_PATH"]
    if os.path.exists(profile_path):
        data = toml.load(profile_path)
    else:
        data = {}
    # Set active profile
    profile_name = config["PROFILE_NAME"]
    data["active"] = profile_name
    # Set config under [profiles.<profile_name>]
    if "profiles" not in data:
        data["profiles"] = {}
    data["profiles"][profile_name] = {
        "PREFECT_API_URL": config["SERVER_URL"],
        "PREFECT_HOME": config["PREFECT_HOME"],
        "PREFECT_API_DATABASE_CONNECTION_URL": config["DB_URL"],
    }
    with open(profile_path, "w") as f:
        toml.dump(data, f)
    print(f"✅ Prefect profile '{profile_name}' written to {profile_path}")


def prefect_server_status():
    """
    Get prefect server status
    """
    config = prefect_config()
    try:
        with socket.create_connection(
            (config["SERVER_HOST"], config["SERVER_PORT"]), timeout=2
        ):
            return True
    except OSError:
        return False


def get_prefect_env():
    """
    Get environment variables of prefect
    """
    config = prefect_config()
    env = os.environ.copy()
    env["PREFECT_HOME"] = config["PREFECT_HOME"]
    env["PREFECT_API_MODE"] = "server"
    env["PREFECT_API_DATABASE_CONNECTION_URL"] = config["DB_URL"]
    env["PREFECT_SERVER_ALLOW_EPHEMERAL_MODE"] = "false"
    env["PREFECT_API_URL"] = config["SERVER_URL"]
    env["PREFECT_PROFILE"] = config["PROFILE_NAME"]
    env["PREFECT_PROFILES_PATH"] = config["PROFILE_PATH"]
    env["PREFECT_LOCAL_STORAGE_PATH"] = config["STORAGE"]
    env["PREFECT_LOGGING_SETTINGS_PATH"] = config["LOGGING_PATH"]
    env["PREFECT_MEMO_STORE_PATH"] = config["MEMO_PATH"]
    return env


def save_prefect_env_to_file():
    """
    Save current Prefect server env config to a .env file for reuse.
    """
    config = prefect_config()
    cachedir = config["CACHEDIR"]
    with open(config["ENV_FILE"], "w") as f:
        f.write(f"PREFECT_HOME={config['PREFECT_HOME']}\n")
        f.write("PREFECT_API_MODE=server\n")
        f.write(f"PREFECT_API_DATABASE_CONNECTION_URL={config['DB_URL']}\n")
        f.write("PREFECT_SERVER_ALLOW_EPHEMERAL_MODE=false\n")
        f.write(f"PREFECT_API_URL={config['SERVER_URL']}\n")
        f.write(f"PREFECT_PROFILE={config['PROFILE_NAME']}\n")
        f.write(f"PREFECT_PROFILES_PATH={config['PROFILE_PATH']}\n")
        f.write(f"PREFECT_LOCAL_STORAGE_PATH={config['STORAGE']}\n")
        f.write(f"PREFECT_LOGGING_SETTINGS_PATH={config['LOGGING_PATH']}\n")
        f.write(f"PREFECT_MEMO_STORE_PATH={config['MEMO_PATH']}\n")
    print(f"📄 Saved Prefect server environment to {config['ENV_FILE']}")
    if os.path.exists(f"{cachedir}/prefect.dashboard") is not True:
        with open(f"{cachedir}/prefect.dashboard", "w") as f:
            f.write(f"{config['SERVER_DASHBOARD']}")
    write_prefect_profile()


def start_server(show_config=False):
    """
    Start prefect server if it is not running
    """
    config = prefect_config()
    cachedir = config["CACHEDIR"]
    if prefect_server_status():
        print(f"🟢 Prefect server is already running at {config['SERVER_DASHBOARD']}")
        if os.path.exists(f"{cachedir}/prefect.dashboard") is not True:
            with open(f"{cachedir}/prefect.dashboard", "w") as f:
                f.write(f"{config['SERVER_DASHBOARD']}")
        if show_config:
            show_prefect_config()
        os.makedirs(config["PREFECT_HOME"], exist_ok=True)
        save_prefect_env_to_file()
        return 0
    print("🚀 Starting Prefect server...")
    pid_file = config["PID_FILE"]
    if os.path.exists(pid_file):
        stop_prefect_server()
    with open(config["LOG_FILE"], "w") as f:
        server_proc = subprocess.Popen(
            [
                "prefect",
                "server",
                "start",
                "--host",
                config["SERVER_HOST"],
                "--port",
                config["SERVER_PORT"],
            ],
            stdout=f,
            stderr=subprocess.STDOUT,
            env=get_prefect_env(),
        )
    server_started = False
    for _ in range(30):  # wait up to 30s for the server to respond
        if prefect_server_status():
            if show_config:
                show_prefect_config()
            server_started = True
            break
        time.sleep(1)
    if server_started:
        with open(pid_file, "w") as pf:
            pf.write(str(server_proc.pid))
        os.makedirs(config["PREFECT_HOME"], exist_ok=True)
        save_prefect_env_to_file()
        print(f"✅ Prefect server is now running at {config['SERVER_DASHBOARD']}")
        if os.path.exists(f"{cachedir}/prefect.dashboard") is not True:
            with open(f"{cachedir}/prefect.dashboard", "w") as f:
                f.write(f"{config['SERVER_DASHBOARD']}")
        return 0
    else:
        print(f"⚠️ Server did not respond in time. Check logs at {config['LOG_FILE']}")
        return 1


def stop_prefect_server():
    """
    Stop prefect server running in the current installation
    Note: it will only stop prefect server which is running from the current installation
    For this pipeline, default port (4250) is kept seperate from default prefect port 4200.
    """
    config = prefect_config()
    pid_file = config["PID_FILE"]
    cachedir = config["CACHEDIR"]
    try:
        if not os.path.exists(pid_file):
            print("⚠️ No PID file found. Cannot stop Prefect server.")
        else:
            with open(pid_file, "r") as f:
                pid = int(f.read().strip())
            print(f"🛑 Stopping Prefect server with PID {pid} ...")
            os.kill(pid, signal.SIGTERM)
            os.remove(pid_file)
            print("✅ Server stopped and PID file removed.")
    except ProcessLookupError:
        print(f"⚠️ No such process with PID {pid}. Removing stale PID file.")
        os.remove(pid_file)
    except Exception as e:
        print(f"❌ Error stopping server: {e}")
    finally:
        os.remove(config["ENV_FILE"])
        os.system(f"touch {config['ENV_FILE']}")
        os.remove(f"{cachedir}/prefect.dashboard")


def show_prefect_config():
    """
    Print the effective Prefect config in this environment.
    """
    config = prefect_config()
    load_dotenv(dotenv_path=config["ENV_FILE"], override=False)
    env = os.environ.copy()
    print("🔍 Prefect config in current environment ...")
    subprocess.run(["prefect", "config", "view"], env=env)


# Exposing only functions
__all__ = [
    name
    for name, obj in globals().items()
    if isinstance(obj, types.FunctionType) and obj.__module__ == __name__
]



import types
import resource
import psutil
import dask
import numpy as np
import warnings
import gc
import logging
import time
import glob
import os
import subprocess
import sys
import tempfile
import shutil
from pathlib import Path
from dask import delayed, compute, config
from dask.distributed import Client, LocalCluster
from dask_jobqueue import SLURMCluster
from datetime import datetime as dt, timedelta
from .basic_utils import *


#################################
# Process management
#################################
def get_nprocess_solarpipe(jobid):
    """
    Get numbers of processes currently running

    Parameters
    ----------
    workdir : str
        Work directory name
    jobid : int
        Job ID

    Returns
    -------
    int
        Number of running processes
    """
    cachedir = get_cachedir()
    pid_file = f"{cachedir}/pids/pids_{jobid}.txt"
    pids = np.loadtxt(pid_file, unpack=True)
    n_process = 0
    for pid in pids:
        if psutil.pid_exists(int(pid)):
            n_process += 1
    return n_process


def get_jobid():
    """
    Get Job ID with millisecond-level uniqueness.

    Returns
    -------
    int
        Job ID in the format YYYYMMDDHHMMSSmmm (milliseconds)
    """
    cachedir = get_cachedir()
    jobid_file = os.path.join(cachedir, "jobids.txt")
    if os.path.exists(jobid_file):
        prev_jobids = np.loadtxt(jobid_file, unpack=True, dtype="int64")
        if prev_jobids.size == 0:
            prev_jobids = []
        elif prev_jobids.size == 1:
            prev_jobids = [str(prev_jobids)]
        else:
            prev_jobids = [str(jid) for jid in prev_jobids]
    else:
        prev_jobids = []

    if len(prev_jobids) > 0:
        FORMAT = "%Y%m%d%H%M%S%f"
        CUTOFF = dt.utcnow() - timedelta(days=15)
        filtered_prev_jobids = []
        for job_id in prev_jobids:
            job_time = dt.strptime(job_id.ljust(20, "0"), FORMAT)  # pad if truncated
            if job_time >= CUTOFF or job_id == 0:  # Job ID 0 is always kept
                filtered_prev_jobids.append(job_id)
        prev_jobids = filtered_prev_jobids

    now = dt.utcnow()
    cur_jobid = (
        now.strftime("%Y%m%d%H%M%S") + f"{int(now.microsecond/1000):03d}"
    )  # ms = first 3 digits of microseconds
    prev_jobids.append(cur_jobid)

    job_ids_int = np.array(prev_jobids, dtype=np.int64)
    np.savetxt(jobid_file, job_ids_int, fmt="%d")

    return int(cur_jobid)


def save_main_process_info(pid, jobid, msname, workdir, outdir, cpu_frac, mem_frac):
    """
    Save main processes info

    Parameters
    ----------
    pid : int
        Main job process id
    jobid : int
        Job ID
    msname : str
        Main measurement set
    workdir : str
        Work directory
    outdir : str
        Output directory
    cpu_frac : float
        CPU fraction of the job
    mem_frac : float
        Memory fraction of the job

    Returns
    -------
    str
        Job info file name
    """
    cachedir = get_cachedir()
    prev_main_pids = glob.glob(f"{cachedir}/main_pids_*.txt")
    prev_jobids = [
        str(os.path.basename(i).rstrip(".txt").split("main_pids_")[-1])
        for i in prev_main_pids
    ]
    if len(prev_jobids) > 0:
        FORMAT = "%Y%m%d%H%M%S%f"
        CUTOFF = dt.utcnow() - timedelta(days=15)
        filtered_prev_jobids = []
        for i in range(len(prev_jobids)):
            job_id = prev_jobids[i]
            job_time = dt.strptime(job_id.ljust(20, "0"), FORMAT)  # pad if truncated
            if job_time < CUTOFF or job_id == 0:  # Job ID 0 is always kept
                filtered_prev_jobids.append(job_id)
            else:
                os.system(f"rm -rf {prev_main_pids[i]}")
                if os.path.exists(f"{cachedir}/pids/pids_{job_id}.txt"):
                    os.system(f"rm -rf {cachedir}/pids/pids_{job_id}.txt")
    main_job_file = f"{cachedir}/main_pids_{jobid}.txt"
    main_str = f"{jobid} {pid} {msname} {workdir} {outdir} {cpu_frac} {mem_frac}"
    with open(main_job_file, "w") as f:
        f.write(main_str)
    return main_job_file


def save_pid(pid, pid_file):
    """
    Save PID

    Parameters
    ----------
    pid : int
        Process ID
    pid_file : str
        File to save
    """
    if os.path.exists(pid_file):
        pids = np.loadtxt(pid_file, unpack=True, dtype="int")
        pids = np.append(pids, pid)
    else:
        pids = np.array([int(pid)])
    np.savetxt(pid_file, pids, fmt="%d")


def generate_activate_env(outfile="activate_env.sh"):
    """
    Generate a shell script that activates the current Python environment.

    This works for both Conda and virtualenv environments and is safe for use in
    non-interactive shells (e.g., Slurm batch jobs) by explicitly sourcing `conda.sh`.

    If conda is not found in $PATH, it will try loading either `anaconda` or `anaconda3` module.

    Parameters
    ----------
    outfile : str
        Path to the shell script to write (default: ./activate_env.sh).

    Returns
    -------
    str
        Output file name
    """
    outfile = Path(outfile).expanduser().resolve()
    putfile = os.path.abspath(outfile)
    lines = ["#!/bin/bash", ""]

    def module_exists(name):
        """Check if a module exists using 'module avail'."""
        try:
            subprocess.run(
                ["module", "avail", name],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=False,
            )
            return True
        except Exception:
            return False

    # Conda-based environment
    if "CONDA_DEFAULT_ENV" in os.environ:
        conda_env = os.environ["CONDA_DEFAULT_ENV"]
        lines.append("# === Activate Conda Environment Safely ===")
        lines.append("if ! command -v conda >/dev/null 2>&1; then")
        if module_exists("anaconda"):
            lines.append("    module load anaconda")
        elif module_exists("anaconda3"):
            lines.append("    module load anaconda3")
        else:
            lines.append("    echo 'No Conda module found (anaconda or anaconda3)'")
            lines.append("    exit 1")
        lines.append("fi")
        lines.append("source $(conda info --base)/etc/profile.d/conda.sh")
        lines.append(f"conda activate {conda_env}")
    # Virtualenv-based environment
    elif "VIRTUAL_ENV" in os.environ:
        venv_path = os.environ["VIRTUAL_ENV"]
        lines.append("# === Activate Virtualenv ===")
        lines.append(f"source {venv_path}/bin/activate")
    else:
        python_path = sys.executable
        lines.append(
            "# === No Conda/Virtualenv Detected — Using current Python directly ==="
        )
        lines.append(f"echo 'No Conda or virtualenv detected; using: {python_path}'")
        lines.append(f"export PATH={os.path.dirname(python_path)}:$PATH")
    # Write file
    with open(outfile, "w") as f:
        f.write("\n".join(lines) + "\n")
    os.chmod(outfile, 0o755)
    print(f"Created activation script at: {outfile}")
    return outfile


def get_total_worker(cluster):
    """
    Get total workers in the cluster

    Parameters
    ----------
    cluster : dask.cluster
        Dask cluster

    Returns
    -------
    int
        Number of workers
    """
    return len(cluster.workers)


def scale_worker_and_wait(dask_cluster, nworker, timeout=60, poll_interval=1):
    """
    Scale worker and wait until it is done

    Parameters
    ----------
    dask_cluster : dask.cluster
        Dask cluster
    nworker : int
        Number of worker
    timeout : float, optional
        Timeout, show a warning and move
    poll_interval : float, optional
        Check interval in seconds
    """
    print(f"Start scaling to {nworker} workers")
    dask_cluster.scale(nworker)
    timeout = 60
    c = 0
    while c < timeout:
        if get_total_worker(dask_cluster) == nworker:
            print(f"Successfully scaled to {nworker} workers")
            return 0
        else:
            time.sleep(poll_interval)
            c += poll_interval
    print(f"Dask cluster did not scale to {nworker} within {timeout} seconds.")
    return 1


def wait_for_dask_workers(client, min_worker=1, timeout=60):
    """
    Wait until the Dask cluster has a minimum number of total and/or new workers.

    Parameters
    ----------
    client : dask.distributed.Client
        Dask client
    min_worker : int, optional
        Minimum new connected workers (default: 1)
    timeout : float, optional
        Maximum time to wait in seconds (default: 60)

    Raises
    ------
    TimeoutError
        If the required number of workers do not connect in time.
    """
    client.wait_for_workers(n_workers=min_worker, timeout=timeout)


def get_scheduler_name():
    """
    Get job scheduler available

    Returns
    -------
    str
        Scheduler name (local, pbs, slurm)
    """
    if shutil.which("sbatch"):
        return "slurm"
    elif shutil.which("bsub"):
        return "lsf"
    elif shutil.which("qhost"):
        return "sge"
    elif shutil.which("qsub"):
        return "pbs"
    elif shutil.which("condor_submit"):
        return "htcondor"
    elif shutil.which("msub"):
        return "mab"
    elif shutil.which("oarsub"):
        return "oar"
    else:
        return "local"


def get_local_dask_cluster(
    njobs,
    dask_dir,
    cpu_frac=0.8,
    mem_frac=0.8,
    ncpu=-1,
    mem=-1,
    spill_frac=0.7,
    verbose=True,
):
    """
    Create a local Dask cluster

    Parameters
    ----------
    njobs : int
        Number of MS tasks (ideally = number of MS files)
    dask_dir : str
        Dask temporary directory
    cpu_frac : float, optional
        Fraction of total CPUs to use
    mem_frac : float, optional
        Fraction of total memory to use
    ncpu : int, optional
        Number of CPUs to use (if specified, cpu_frac will be ignored)
    mem : float, optional
        Memory in GB to use (if specified, mem_frac will be ignored)
    spill_frac : float, optional
        Spill to disk at this fraction
    verbose : bool, optional
        Verbose (details of cluster)

    Returns
    -------
    client : dask.distributed.Client
        Dask client
    cluster : dask.distributed.LocalCluster
        Dask cluster
    str
        Dask directory
    """
    logging.getLogger("distributed").setLevel(logging.ERROR)
    print("Creating local cluster on the current node.")
    # Set up Dask working directories
    dask_dir = os.path.join(dask_dir.rstrip("/"), f"dask_{int(time.time())}")
    dask_dir_tmp = os.path.join(dask_dir, "tmp")
    os.makedirs(dask_dir_tmp, exist_ok=True)

    total_cpus = psutil.cpu_count(logical=True)
    total_mem = psutil.virtual_memory().total / 1024**3  # In GB

    # Override fractions if ncpu or mem is provided
    if ncpu > 0:
        cpu_frac = min(ncpu / total_cpus, 0.8)
    if mem > 0:
        mem_frac = min(mem / total_mem, 0.8)
    cpu_frac = min(cpu_frac, 0.8)
    mem_frac = min(mem_frac, 0.8)
    usable_mem = total_mem * mem_frac
    usable_cpus = int(total_cpus * cpu_frac)
    n_workers = max(1, usable_cpus)

    # Raise file descriptor limit
    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    if soft < int(hard * 0.8):
        resource.setrlimit(resource.RLIMIT_NOFILE, (int(hard * 0.8), hard))

    dask.config.set(
        {
            "temporary-directory": dask_dir,
            "distributed.worker.memory.target": spill_frac,
            "distributed.worker.memory.spill": spill_frac + 0.1,
            "distributed.worker.memory.pause": spill_frac + 0.2,
            "distributed.worker.memory.terminate": spill_frac + 0.25,
        }
    )

    cluster = LocalCluster(
        n_workers=1,
        threads_per_worker=1,
        memory_limit=f"{usable_mem}GB",
        local_directory=dask_dir,
        dashboard_address=":0",
        processes=True,
        env={
            "TMPDIR": dask_dir_tmp,
            "TMP": dask_dir_tmp,
            "TEMP": dask_dir_tmp,
            "DASK_TEMPORARY_DIRECTORY": dask_dir_tmp,
            "MALLOC_TRIM_THRESHOLD_": "0",
            "PYTHONWARNINGS": "ignore::UserWarning:contextlib",
        },
    )
    client = Client(cluster, heartbeat_interval="5s")
    client.run_on_scheduler(gc.collect)
    if verbose:
        print("####################################################")
        print(f"Dask dashboard available at: {client.dashboard_link}")
        print("####################################################")
    return client, cluster, dask_dir


def get_slurm_dask_cluster(
    njobs,
    config_yaml,
    dask_dir,
    cpu_frac=0.8,
    mem_frac=0.8,
    ncpu=-1,
    mem=-1,
    spill_frac=0.7,
    verbose=True,
):
    """
    Launch a SLURMCluster using a YAML configuration and return a connected Dask client.

    Parameters
    ----------
    njobs : int
        Number of expected tasks (used for worker scaling)
    config_yaml : str
        Path to Dask SLURMCluster YAML configuration
    dask_dir : str
        Dask working directory (for temporary files)
    cpu_frac : float
        Fraction of total CPUs to use (ignored if ncpu > 0)
    mem_frac : float
        Fraction of total RAM to use (ignored if mem > 0)
    ncpu : int
        Total CPUs to use (overrides cpu_frac)
    mem : float
        Total memory (in GB) to use (overrides mem_frac)
    spill_frac : float
        Fraction of memory to spill to disk
    verbose : bool
        Print Dask dashboard URL and diagnostics

    Returns
    -------
    client : dask.distributed.Client
        Connected Dask client
    cluster : dask_jobqueue.SLURMCluster
        SLURM Dask cluster
    str
        Dask directory used
    """
    logging.getLogger("distributed").setLevel(logging.ERROR)

    dask_dir = os.path.join(dask_dir.rstrip("/"), f"dask_{int(time.time())}")
    dask_dir_tmp = os.path.join(dask_dir, "tmp")
    os.makedirs(dask_dir_tmp, exist_ok=True)

    total_cpus = psutil.cpu_count(logical=True)
    total_mem = psutil.virtual_memory().total / 1024**3  # in GB

    if ncpu > 0:
        cpu_frac = min(ncpu / total_cpus, 0.8)
    if mem > 0:
        mem_frac = min(mem / total_mem, 0.8)

    cpu_frac = min(cpu_frac, 0.8)
    mem_frac = min(mem_frac, 0.8)
    usable_mem = total_mem * mem_frac
    usable_cpus = int(total_cpus * cpu_frac)

    # Raise file descriptor limit
    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    if soft < int(hard * 0.8):
        resource.setrlimit(resource.RLIMIT_NOFILE, (int(hard * 0.8), hard))

    dask.config.set(
        {
            "temporary-directory": dask_dir,
            "distributed.worker.memory.target": spill_frac,
            "distributed.worker.memory.spill": spill_frac + 0.1,
            "distributed.worker.memory.pause": spill_frac + 0.2,
            "distributed.worker.memory.terminate": spill_frac + 0.25,
        }
    )

    # Load cluster config from YAML
    with open(config_yaml, "r") as f:
        cluster_config = yaml.safe_load(f)
    dask.config.update(cluster_config, priority="new")

    # Initialize SLURM cluster
    cluster = SLURMCluster(
        local_directory=dask_dir_tmp,
        env_extra=[
            f"TMPDIR={dask_dir_tmp}",
            f"TMP={dask_dir_tmp}",
            f"TEMP={dask_dir_tmp}",
            f"DASK_TEMPORARY_DIRECTORY={dask_dir_tmp}",
            "MALLOC_TRIM_THRESHOLD_=0",
            "PYTHONWARNINGS=ignore::UserWarning:contextlib",
        ],
    )

    # Scale workers (1 per task/MS file ideally)
    cluster.scale(njobs)
    client = Client(cluster, heartbeat_interval="5s")
    client.run_on_scheduler(gc.collect)
    if verbose:
        print("####################################################")
        print(f"Dask dashboard available at: {client.dashboard_link}")
        print("####################################################")

    return client, cluster, dask_dir


# Exposing only functions
__all__ = [
    name
    for name, obj in globals().items()
    if isinstance(obj, types.FunctionType) and obj.__module__ == __name__
]



import logging
import dask
import numpy as np
import argparse
import traceback
import time
import sys
import os
from dask import delayed
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.ERROR)
logging.getLogger("tornado.application").setLevel(logging.CRITICAL)
datadir = get_datadir()


def cor_sidereal_motion(
    mslist,
    dask_client,
    workdir,
):
    """
    Perform sidereal motion correction

    Parameters
    ----------
    mslist : list
        Measurement set list
    dask_client : dask.client
        Dask client
    workdir : str
        Work directory

    Returns
    -------
    int
        Success message
    list
        List of sidereal motion corrected measurement sets
    """
    try:
        container_name = "meerwsclean"
        container_present = check_udocker_container(container_name)
        if not container_present:
            container_name = initialize_wsclean_container(name=container_name)
            if container_name is None:
                print(
                    "Container {container_name} is not initiated. First initiate container and then run."
                )
                return 1, []

        tasks = []
        for ms in mslist:
            tasks.append(delayed(correct_solar_sidereal_motion)(ms))
        print("Starting sidereal motion correction jobs...")
        results = list(dask_client.gather(dask_client.compute(tasks)))

        splited_ms_list_phaserotated = []
        for i in range(len(results)):
            msg = results[i]
            ms = mslist[i]
            if msg == 0:
                if os.path.exists(ms + "/.sidereal_cor"):
                    splited_ms_list_phaserotated.append(ms)
        if len(splited_ms_list_phaserotated) == 0:
            print("##################")
            print(
                "Sidereal motion correction is not successful for any measurement set."
            )
            print("##################")
            return 1, []
        else:
            print("##################")
            print("Sidereal motion corrections are done successfully.")
            print("##################")
            return 0, splited_ms_list_phaserotated
    except Exception as e:
        traceback.print_exc()
        print("##################")
        print("Sidereal motion correction is not successful for any measurement set.")
        print("##################")
        return 1, []


def main(
    mslist,
    workdir="",
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid=0,
    start_remote_log=False,
    dask_client=None,
):
    """
    Run a parallel processing pipeline for solar sidereal motion correction

    Parameters
    ----------
    mslist : str
        Comma-separated list of paths to measurement sets to be processed.
    workdir : str, optional
        Directory for logs, intermediate files, and other outputs.
        If empty, defaults to the directory of the first MS with `/workdir` appended.
        Default is "".
    cpu_frac : float, optional
        Fraction of total CPU cores to allocate per task. Default is 0.8.
    mem_frac : float, optional
        Fraction of total system memory to allocate per task. Default is 0.8.
    logfile : str or None, optional
        Path to the log file for capturing logs. If None, logging to file is disabled. Default is None.
    jobid : int, optional
        Unique job identifier used for PID tracking and task differentiation. Default is 0.
    start_remote_log : bool, optional
        Whether to enable remote logging based on credentials stored in the workdir. Default is False.
    dask_client : dask.client, optional
        Dask client

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    mslist = mslist.split(",")
    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(mslist[0])) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "do_sidereal_cor", logfile, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")

    dask_cluster = None
    if dask_client is None:
        dask_client, dask_cluster, dask_dir = get_local_dask_cluster(
            2,
            dask_dir=workdir,
            cpu_frac=cpu_frac,
            mem_frac=mem_frac,
        )
        nworker = max(2, int(psutil.cpu_count() * cpu_frac))
        scale_worker_and_wait(dask_cluster, nworker)

    try:
        if len(mslist) == 0:
            print("Please provide a list of measurement sets.")
            msg = 1
        else:
            msg, final_target_mslist = cor_sidereal_motion(
                mslist,
                dask_client,
                workdir,
            )
    except Exception as e:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        for ms in mslist:
            drop_cache(ms)
        drop_cache(workdir)
        clean_shutdown(observer)
        if dask_cluster is not None:
            dask_client.close()
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Correct measurement sets for sidereal motion",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "mslist",
        type=str,
        help="Comma-separated list of measurement sets (required positional argument)",
    )
    basic_args.add_argument("--workdir", type=str, default="", help="Working directory")

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced calibration and imaging parameters\n###################"
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="CPU fraction to use",
        metavar="Float",
    )
    hard_args.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Memory fraction to use",
        metavar="Float",
    )
    hard_args.add_argument("--logfile", type=str, default=None, help="Log file")
    hard_args.add_argument("--jobid", type=int, default=0, help="Job ID")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        return 1

    args = parser.parse_args()

    msg = main(
        mslist=args.mslist,
        workdir=args.workdir,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nSidereal motion corrections are done.\n###################\n"
    )
    os._exit(result)
